program MethodBindingExample;
  { illustrates various method binding alternatives }

uses
  SysUtils; { to raise exceptions for runtime errors }

const
  W = 21; { width of TChild output }

type
  { Each method below writes m and the method's name. }
  { They differ in the binding directives in the parent class: }
  {   S = Static, D = Dynamic, V = Virtual, A = Abstract }
  { and the override directives in the child class: }
  {   I = Inherit, H = Hide, R = Override }
  TParent = class(TObject)
  public
    procedure SI(const m: string);
    procedure SH(const m: string);
    procedure SR(const m: string);
    procedure DI(const m: string); dynamic;
    procedure DH(const m: string); dynamic;
    procedure DR(const m: string); dynamic;
    procedure VI(const m: string); virtual;
    procedure VH(const m: string); virtual;
    procedure VR(const m: string); virtual;
    procedure AI(const m: string); virtual; abstract;
    procedure AH(const m: string); virtual; abstract;
    procedure AR(const m: string); virtual; abstract;
  end; { class TParent }

  TChild = class(TParent)
{   procedure SI(...); inherited }
    procedure SH(const m: string);
{   procedure SR(...); override; } (* SYNTAX ERROR *)
{   procedure DI(...); inherited }
    procedure DH(const m: string); (* COMPILER WARNING *)
    procedure DR(const m: string); override;
{   procedure VI(...); inherited }
    procedure VH(const m: string); (* COMPILER WARNING *)
    procedure VR(const m: string); override;
{   procedure AI(...); inherited }
    procedure AH(const m: string);
    procedure AR(const m: string); override;
  end; { class TChild }

// TParent method implementations

procedure TParent.SI(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TParent.SI')
  end; { TParent.SI }

procedure TParent.SH(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TParent.SH')
  end; { TParent.SH }

procedure TParent.SR(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TParent.SR')
  end; { TParent.SR }

procedure TParent.DI(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TParent.DI')
  end; { TParent.DI }

procedure TParent.DH(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TParent.DH')
  end; { TParent.DH }

procedure TParent.DR(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TParent.DR')
  end; { TParent.DR }

procedure TParent.VI(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TParent.VI')
  end; { TParent.VI }

procedure TParent.VH(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TParent.VH')
  end; { TParent.VH }

procedure TParent.VR(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TParent.VR')
  end; { TParent.VR }

{ procedure TParent.AI; cannot be implemented }

{ procedure TParent.AH; cannot be implemented }

{ procedure TParent.AR; cannot be implemented }

// TChild method implementations

procedure TChild.SH(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TChild.SH' : W)
  end; { TChild.SH }

{ procedure TChild.SR;  syntactically disallowed }

procedure TChild.DH(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TChild.DH' : W)
  end; { TChild.DH }

procedure TChild.DR(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TChild.DR' : W)
  end; { TChild.DR }

procedure TChild.VR(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TChild.VR' : W)
  end; { TChild.VR }

procedure TChild.AR(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TChild.AR' : W)
  end; { TChild.AR }

procedure TChild.VH(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TChild.VH' : W)
  end; { TChild.VH }

procedure TChild.AH(const m: string);
  begin
    WriteLn(m, ' invokes ', 'TChild.AH' : W)
  end; { TChild.AH }

var
  pp, pc: TParent; cp, cc: TChild;

begin
  pp := TParent.Create  (* COMPILER WARNING *)
; pc := TChild.Create  (* COMPILER WARNING *)
(* ; cp := TParent.Create  COMPILER ERROR *)
; cc := TChild.Create  (* COMPILER WARNING *)

; pp.SI('pp.SI')
; pc.SI('pc.SI')
; cc.SI('cc.SI')

; pp.SH('pp.SH')
; pc.SH('pc.SH')
; cc.SH('cc.SH')

; pp.DI('pp.DI')
; pc.DI('pc.DI')
; cc.DI('cc.DI')

; pp.DH('pp.DH')
; pc.DH('pc.DH')
; cc.DH('cc.DH')

; pp.DR('pp.DR')
; pc.DR('pc.DR')
; cc.DR('cc.DR')

; pp.VI('pp.VI')
; pc.VI('pc.VI')
; cc.VI('cc.VI')

; pp.VH('pp.VH')
; pc.VH('pc.VH')
; cc.VH('cc.VH')

; pp.VR('pp.VR')
; pc.VR('pc.VR')
; cc.VR('cc.VR')

; try pp.AI('pp.AI')
  except on e : Exception do
    WriteLn('pp.AI raises exception: ', e.Message)
  end { try }
; try pc.AI('pc.AI')
  except on e : Exception do
    WriteLn('pc.AI raises exception: ', e.Message)
  end { try }
; try cc.AI('cc.AI')
  except on e : Exception do
    WriteLn('cc.AI raises exception: ', e.Message)
  end { try }

; try pp.AH('pp.AH')
  except on e : Exception do
    WriteLn('pp.AH raises exception: ', e.Message)
  end { try }
; try pc.AH('pp.AH')
  except on e : Exception do
    WriteLn('pc.AH raises exception: ', e.Message)
  end { try }
; cc.AH('cc.AH')

; try pp.AR('pp.AR')
  except on e : Exception do
    WriteLn('pp.AR raises exception: ', e.Message)
  end { try }
; pc.AR('pc.AR')
; cc.AR('cc.AR')

end.
